from .docs import _auto_fix_docs
from .fields import FilterField
from .filter_set import FilterSet, create_filters_from_set
from .filters import create_filters, create_filters_from_model
from .operators import FilterOperator
from .sorters import create_sorting, create_sorting_from_model
from .types import FiltersResolver, FilterValues, SortingResolver, SortingValues

# Auto-patch FastAPI to fix explode parameter
_auto_fix_docs()

__all__ = [
    "FilterField",
    "FilterOperator",
    "FilterSet",
    "FilterValues",
    "FiltersResolver",
    "SortingResolver",
    "SortingValues",
    "create_filters",
    "create_filters_from_model",
    "create_filters_from_set",
    "create_sorting",
    "create_sorting_from_model",
]
